# User Guide: Debugging Programs on SHMAC 
This is a guide that cover the necessary setup for using GDB for SHMAC. Note that it does not explain how to use GDB in general. 

The debugger is compatible with any SHMAC configuration having the following properties:
* ARMv3 compatible CPU cores (e.g. *arv3*).
* The updated APB tile with 16 TTY channels.

## Install GDB 7.7 for ARM
The stub is designed for GDB 7.7 or newer. Older versions are not tested, but might still work. The GDB installation must have ARM as target platform. The easiest way to obtain the correct version is to compile GDB from source. You may install GDB either on your computer or on the host controller on SHMAC.

1. Download and extract the source code (http://ftp.gnu.org/gnu/gdb/gdb-7.7.tar.gz).
2. Compile GDB using the following commands:

```sh
mkdir gdb-build
cd gdb-build
CFLAGS=-Wno-error=deprecated-declarations [path to source code]/configure --target=armv3-none-eabi
make
```

## Setup *ser2net* on SHMAC Host Controller
This step is not necessary if you intent to run GDB on host controller.

1. Install ser2net from package manager:
```sh
sudo apt-get install ser2net
```
2. Edit the file **/etc/ser2net.conf** and remove the default configuration. Add the following lines to the bottom of the file:
```config
4242:raw:600:/dev/ttySHMAC0:115200 8DATABITS NONE 1STOPBIT
4243:raw:600:/dev/ttySHMAC1:115200 8DATABITS NONE 1STOPBIT
4244:raw:600:/dev/ttySHMAC2:115200 8DATABITS NONE 1STOPBIT
4245:raw:600:/dev/ttySHMAC3:115200 8DATABITS NONE 1STOPBIT
4246:raw:600:/dev/ttySHMAC4:115200 8DATABITS NONE 1STOPBIT
4247:raw:600:/dev/ttySHMAC5:115200 8DATABITS NONE 1STOPBIT
4248:raw:600:/dev/ttySHMAC6:115200 8DATABITS NONE 1STOPBIT
4249:raw:600:/dev/ttySHMAC7:115200 8DATABITS NONE 1STOPBIT
4250:raw:600:/dev/ttySHMAC8:115200 8DATABITS NONE 1STOPBIT
4251:raw:600:/dev/ttySHMAC9:115200 8DATABITS NONE 1STOPBIT
4252:raw:600:/dev/ttySHMAC10:115200 8DATABITS NONE 1STOPBIT
4253:raw:600:/dev/ttySHMAC11:115200 8DATABITS NONE 1STOPBIT
4254:raw:600:/dev/ttySHMAC12:115200 8DATABITS NONE 1STOPBIT
4255:raw:600:/dev/ttySHMAC13:115200 8DATABITS NONE 1STOPBIT
4256:raw:600:/dev/ttySHMAC14:115200 8DATABITS NONE 1STOPBIT
4257:raw:600:/dev/ttySHMAC15:115200 8DATABITS NONE 1STOPBIT
```
This configuration will map */dev/ttySHMAC0* to IP port 4242, */dev/ttySHMAC1* to 4243, etc.
3. Restart the ser2net service using the following command:
```sh
sudo /etc/init.d/ser2net restart
```

## Compile the GDB Stub
The GDB stub + test application source code is located at https://bitbucket.org/bjorncs/gdb-project. The source code can be downloaded manually from website or by using git:

```sh
git clone https://bjorncs@bitbucket.org/bjorncs/gdb-project.git
```

Prerequisites:
* A GCC toolchain for ARMv3. The GDB stub needs the *libc* headers. The test program requires the *libc* itself in addition to the headers. The repository contains a guide for setting up the toolchain required for compiling the GDB stub and Barrelfish.

The header file **defs.h** must be configured prior to compilation. It's located in the directory **gdb-stub/include**. Modify the definitions defining the number of cores and TTY channels:

```C
#pragma once

#define NUM_CORES 3
#define NUM_TTY_CHANNELS 16
```

Navigate to the source code root directory. You may build the test program and the stub with *make*. If you only want to build the stub, run *make* from the directory **gdb-stub**. The resulting static library is (**libgdbstub.a**) located in **gdb-stub/build/libs/**. 

## Integrate GDB Stub with Program
### Insert Interrupt Hooks

1. Modify the interrupt vector so that the the GDB stub handles irq interrupts and undefined instruction exceptions:
```asm
vector:
    b reset_handler
    b gdb_undef_instr_handler
    b swi_handler
    b . // Prefetch abort
    b . // Data abort
    b . // Reserved
    b gdb_irq_handler
    b . // FIQ
```
2. Label the entry point for the program's irq handler and undefined instruction handler with gdb_irq_handler_program and gdb_undef_instr_handler_program respectively:
```asm
gdb_irq_handler_program:
irq_handler:
    // Save regs
    push {r0-r3,r12,lr}
    ...

gdb_undef_instr_handler_program:
undef_instr_handler:
    // Save regs
    push {r0-r3,r12,lr}
    ...
```
### Insert Call to *gdb_init*
Insert a call to the *gdb_init* function at the start of program. It takes a *gdb_config_t* struct as only argument. *gdb_config_t* is defined in header file **gdb.h**. Example:

```C
#include "gdb.h"
...
void die(void) {
    ...
}
...
int main(void) {
    int cpu_id = *TILEREG_CPUID;
    if (cpu_id == 0) {

        gdb_config_t gdb_config;
        gdb_config.tty_channel[0] = 3;
        gdb_config.tty_channel[1] = 4;
        gdb_config.tty_channel[2] = 5;
        gdb_config.kill_program_handler = &die;
        gdb_init(&gdb_config);
    }
    ...
}
```

The config above configures core 0, 1 and 2 to use TTY channel 3, 4 and 5 respectively.
The function *die* will be invoked by the debugger when killing the program. It's up to the program to decide what this function should do.

IMPORTANT: Make sure that *gdb_init* is called by a single core only (in case the main function is run by multiple cores).

### Insert the Initial Static Breakpoint

Place a static breakpoint somewhere after the call to *gdb_init*. The static breakpoint is defined as a C pre-processor macro (*GDB_TRAP*). The program will initially halt at the location of this breakpoint.
```C
...
GDB_TRAP;
...
```

### Compile the Program with the GDB Stub
Compile the program with the static library **libgdbstub.a**. The necessary header files are located at **gdb-stub/include**. Make sure to compile with debug symbols. It's recommended to compile without optimizations.

## Starting a Debugging Session

1. Start the program on SHMAC using *shmac_program* and *shmac_reset*.
2. Launch GDB on the host machine. Type the following commands to the GDB CLI to attach to the program running on SHMAC:
```gdb
set remotetimeout 12
set remote trace-status-packet 0
set remote query-attached-packet 0
set can-use-hw-watchpoints 0
file [path to program elf file]
target remote [shmac host ip]:[shmac host port]
```
3. GDB is now ready to debug the program. 


## Using the Debugger
### Interrupting the Program
You may halt the program by pressing Ctrl+C in the GDB CLI.
### Static Breakpoints
You may insert static breakpoints to the program as described earlier.
### Disabling the Debugger
You may disable the debugger by compiling the stub (**libgdbstub.a**) with the flag *-D DISABLE_DEBUGGER*. Another way of disabling the debugger is to disable each core individually in the *gdb_config* struct, as shown earlier.
### Printing the SPSR Register
You can print the content of the SPSR register with the following command:
```gdb
p $fps
```
