#include <string.h>
#include "breakpoint.h"
#include "misc.h"
#include "mutex.h"

static volatile uint32_t mutex;

#define MAX_BREAKPOINTS 32

typedef volatile struct {
    uint32_t *addr;
    uint32_t original_instruction;
    bool valid;
} breakpoint_t;

typedef volatile struct {
    breakpoint_t breakpoints[MAX_BREAKPOINTS];
    breakpoint_t step_breakpoint;
} core_state_t;

static core_state_t state[NUM_CORES];

// Check if there is a breakpoint at specified address
static bool is_breakpoint(uint32_t * addr);

// Misc
static unsigned count_bits(int n);
static bool should_execute(uint8_t status_bits, uint8_t cond_bits);
static uint32_t shift_value(uint8_t shift_field, uint32_t value, uint32_t * registers);
static uint32_t rotate(uint32_t value, uint32_t shift);

void breakpoint_init(void)
{
    gdb_init_mutex(&mutex);
    for (int i = 0; i < NUM_CORES; i++)
    {
        core_state_t * s = &state[i];
        s->step_breakpoint.valid = false;
        s->breakpoints[0].valid = false;
    }

}

void breakpoint_clear_trap()
{
    gdb_lock_mutex(&mutex);
    unsigned core_id = get_cpu_id();
    breakpoint_t * bp = &(state[core_id].step_breakpoint);
    if (bp->valid)
    {
        uint32_t *addr = bp->addr;
        // Make sure its a step breakpoint or interrupt breakpoint
        if (*addr == STEP_BREAKPOINT_INSTR || *addr == INTERRUPT_BREAKPOINT_INSTR)
        {
           *(bp->addr) = bp->original_instruction;
        }
        bp->valid = false;
    }
    gdb_unlock_mutex(&mutex);
}
// Note: Single stepping is not guaranteed to work. Will corrupt the execution state in worst case.
// Catches the typical use case 
void breakpoint_single_step(uint32_t *registers, uint32_t *addr)
{
    gdb_lock_mutex(&mutex);
    // This is the default case for single step breakpoint location
    uint32_t* next_instr = addr + 1;

    // First check condition bits
    uint8_t cond_bits = *addr >> 28;
    uint8_t status_bits = registers[CPSR] >> 28;

    gdb_debug_printf(buffer, "SS: Status bits %#x, cond_bits %#x, should_execute: %d\n", status_bits, cond_bits, should_execute(status_bits, cond_bits));
    
    // The instruction will be executed, lets check if it does something with PC
    if (should_execute(status_bits, cond_bits))
    {
        if (is_mov_pc_instr(*addr))
        {
            next_instr = (uint32_t *) handle_mov_pc_instr(*addr, registers);
        } else if (is_ldm_instr(*addr))
        {
            next_instr = (uint32_t *) handle_ldm_instr(*addr, registers);
        } else if (is_branch_instr(*addr))
        {
            next_instr = (uint32_t *) handle_branch_instr(addr);
        } else if (is_sub_pc_instr(*addr))
        {
            next_instr = (uint32_t *) handle_sub_pc_instr(*addr, registers);
        } else if (is_add_pc_instr(*addr))
        {
            next_instr = (uint32_t *) handle_add_pc_instr(*addr, registers);
        } else if (is_swi_instr(*addr))
        {
            next_instr = (uint32_t *) handle_swi_instr(); 
        } else if (is_ldr_pc_instr(*addr))
        {
            next_instr = (uint32_t *) handle_ldr_pc_instr(*addr, registers);
        }
    }
    // No point inserting single step breakpoint if there is already a breakpoint there
    if (!is_breakpoint(next_instr))
    {
        unsigned core_id = get_cpu_id();
        breakpoint_t * bp = &(state[core_id].step_breakpoint);

        bp->addr = next_instr;
        bp->original_instruction = *next_instr;
        bp->valid = true;
        *next_instr = STEP_BREAKPOINT_INSTR;
        gdb_debug_printf(buffer, "SS: Step breakpoint: %#x\n", bp->addr);
    }
    gdb_unlock_mutex(&mutex);
}

bool breakpoint_insert(uint32_t *addr)
{
    gdb_lock_mutex(&mutex);
    // Breakpoints already inserted. 
    // GDB docs state that this operation should be implemented in a idempotent way.
    if (is_breakpoint(addr))
    {
        gdb_unlock_mutex(&mutex);
        return true;
    }
    
    // Search for empty slot
    unsigned core_id = get_cpu_id();
    breakpoint_t * breakpoints = &(state[core_id].breakpoints[0]);

    int index = 0;
    for (; index < MAX_BREAKPOINTS && breakpoints[index].valid; index++);
    if (index == MAX_BREAKPOINTS)
    {
        // Too many breakpoints
        gdb_unlock_mutex(&mutex);
        return false;
    }
    breakpoints[index].addr = addr;
    breakpoints[index].original_instruction = *addr;
    breakpoints[index].valid = true;
    *addr = STANDARD_BREAKPOINT_INSTR;
    gdb_unlock_mutex(&mutex);
    return true;
}

void breakpoint_delete(uint32_t *addr)
{
    gdb_lock_mutex(&mutex);
    unsigned core_id = get_cpu_id();
    breakpoint_t * breakpoints = &(state[core_id].breakpoints[0]);
    
    int index = 0;
    for (;index < MAX_BREAKPOINTS && (breakpoints[index].addr != addr || !breakpoints[index].valid); index++);
    if (index == MAX_BREAKPOINTS)
    {
        gdb_unlock_mutex(&mutex);
        return;
    }
    // Only delete if there is a breakpoint there
    if (*addr == STANDARD_BREAKPOINT_INSTR)
    {
        *addr = breakpoints[index].original_instruction;
    }
    breakpoints[index].valid = false;
    gdb_unlock_mutex(&mutex);
}

void breakpoint_trap(uint32_t *pc)
{
    // Insert a single step breakpoint at location pointed by PC
    breakpoint_clear_trap();

    gdb_lock_mutex(&mutex);
    // No need to trap if there is already a breakpoint there
    if (!is_breakpoint(pc))
    {
        unsigned core_id = get_cpu_id();
        breakpoint_t * bp = &(state[core_id].step_breakpoint);
        bp->addr = pc;
        bp->original_instruction = *pc;
        bp->valid = true;
        *pc = INTERRUPT_BREAKPOINT_INSTR;
    }
    gdb_unlock_mutex(&mutex);
}

bool is_breakpoint(uint32_t * addr)
{
    return *addr == STANDARD_BREAKPOINT_INSTR || *addr == INTERRUPT_BREAKPOINT_INSTR 
            || *addr == STEP_BREAKPOINT_INSTR || *addr == STANDARD_BREAKPOINT_INSTR;
}

// Check if instruciton is a mov{s} pc, xx
// Note: Does not detect shifting of register value to PC
bool is_mov_pc_instr(uint32_t instr)
{
    return (instr & 0xFE0F000) == 0x1A0F000;
}

// Check if the instruction is a LDM (load-multiple) with PC in its register list
// This instruction is typically used by the GCC compiler in the form of a stack pop into PC as return sequence.
bool is_ldm_instr(uint32_t instr)
{
    return (instr & 0xE108000) == 0x8108000;
}

// Check if instruction is a branch - match with pattern 0xXXXX101XXXX...
bool is_branch_instr(uint32_t instr)
{
    return (instr & 0xE000000) == 0xA000000;
}

//Check if the instruction is a sub pc, rr, #xx
bool is_sub_pc_instr(uint32_t instr)
{
    return (instr & 0xFE0F000) == 0x240F000;
}

bool is_add_pc_instr(uint32_t instr)
{
    // Encoding xxxx 00 x 0100 x xxxx 1111 xxxxxxxxxxxx
    // op-code = 0100
    // Mask:    0000 11 0 1111 0 0000 1111 000000000000 = 0x0DE0F000
    // Map:     0000 00 0 0100 0 0000 1111 000000000000 = 0x0080F000
    return (instr & 0x0DE0F000) == 0x0080F000;
}

bool is_swi_instr(uint32_t instr)
{
    return (instr & 0xF000000) == 0xF000000;
}

bool is_ldr_pc_instr(uint32_t instr)
{
    // Encoding xxxx 01 xxxxx1 xxxx 1111 xxxxxxxxxxxx
    // Mask:    0000 11 000001 0000 1111 000000000000 = 0xC10F000
    // Map:     0000 01 000001 0000 1111 000000000000 = 0x410F000
    return (instr & 0xC10F000) == 0x410F000;
}

uint32_t handle_mov_pc_instr(uint32_t instr, uint32_t * registers)
{
    uint8_t source_register = instr & 0xF;
    gdb_debug_printf(buffer, "SS: mov pc, xx - source register %d\n", source_register);
    return registers[source_register];
}

uint32_t handle_ldm_instr(uint32_t instr, uint32_t * registers)
{
    bool pre_index = ((1 << 24) & instr) != 0;
    bool up = ((1 << 23) & instr) != 0;
    int register_bitmap = instr & 0xFFFF;
    int base_register = (instr >> 16) & 0xF;
    int pc_order = count_bits(register_bitmap);

    int offset;
    if (up)
    {
        offset = pre_index ? pc_order : pc_order - 1;
    } else
    {
        offset = pre_index ? -1 : 0;
    }

    uint32_t *pc_location = (uint32_t *) (registers[base_register] + 4 * offset);
    
    gdb_debug_printf(buffer, "SS: ldm/pop instr: %#x\n", instr);
    gdb_debug_printf(buffer, "SS: Base register=%d, pre_index=%d, up=%d, register_bitmap=%d\n", base_register, pre_index, up, register_bitmap);
    gdb_debug_printf(buffer, "SS: pc_order=%d, offset=%d, pc_location=%p\n", pc_order, offset, pc_location);
    
    return (uint32_t) pc_location;
}

uint32_t handle_ldr_pc_instr(uint32_t instr, uint32_t * registers)
{
    bool immediate = ((instr >> 25) & 1) == 0;
    int offset_modifier = ((instr >> 23) & 1) == 1 ? 1 : -1;

    uint8_t base_register = (instr >> 16) & 0xF;
    uint32_t base_address = registers[base_register];
    if (base_register == 15)
    {
        base_address += 8;
    }

    if (immediate)
    {
        uint32_t offset = instr & 0xFFF;
        uint32_t * pc_location = (uint32_t *) (base_address + offset * offset_modifier);
        gdb_debug_printf(buffer, "SS: ldr pc, [r%d, #%d], pc_location = %#x\n", base_register, offset * offset_modifier, pc_location);
        return (uint32_t) pc_location;
    } else
    {
        uint32_t offset_register = instr & 0xF;
        uint32_t offset_value = registers[offset_register];
        uint8_t shift_field = (instr >> 4) & 0xFF;
        uint32_t offset = shift_value(shift_field, offset_value, registers);
        uint32_t * pc_location = (uint32_t *) (base_address + offset * offset_modifier);
        gdb_debug_printf(buffer, "SS: ldr pc, [r%d, r%d, lsl], pc_location = %#x\n", base_register, offset_register, pc_location);
        return (uint32_t) pc_location;
    }
}


uint32_t handle_branch_instr(uint32_t *instr)
{
    int32_t offset = *instr & 0xFFFFFF;  
    // Shift left two bit and sign extend to 32-bit
    offset <<= 8;
    offset /= 64;
    int32_t currentAddr = (int32_t) instr;
    
    // Add offset + 8 for pipeline thingy to current address
    gdb_debug_printf(buffer, "SS: Branch: offset %d, instr %#x\n", offset, *instr);
    
    return (offset + currentAddr + 8);
}

uint32_t handle_sub_pc_instr(uint32_t instr, uint32_t * registers)
{
    int op1_register = (instr >> 16) & 0xF;
    int op1 = registers[op1_register];
    int op2 = instr & 0xFF;
    gdb_debug_printf(buffer, "SS: sub pc, r%d, #%d\n", op1_register, op2);
    return op1 - op2;
}

uint32_t handle_add_pc_instr(uint32_t instr, uint32_t * registers)
{
    bool immediate = ((instr >> 25) & 1) == 1;
    unsigned op1_register = (instr >> 16) & 0xF;
    int op1 = registers[op1_register];
    if (op1_register == R15)
    {
        op1 += 8; // PC points two instructions ahead
    }

    // Check if operand 2 is register or immediate
    if (immediate)
    {
        unsigned imm = instr & 0xFF;
        unsigned rotate_value = (instr >> 8) & 0xF;
        int op2 = rotate(imm, rotate_value);
        gdb_debug_printf(buffer, "SS: add pc, r%d, #%d, rotate %d\n", op1_register, imm, rotate_value);
        return op1 + op2;
    } else
    {
        unsigned op2_register = instr & 0xF;
        int op2_value = registers[op2_register];
        uint8_t shift_field = (instr >> 4) & 0xFF;
        int op2 = shift_value(shift_field, op2_value, registers);
        gdb_debug_printf(buffer, "SS: add pc, r%d, r%d, op1=%#x, op2=%#x\n", op1_register, op2_register, op1, op2);
        return op1 + op2;
    }
}

uint32_t handle_swi_instr(void)
{
    // Insert breakpoint at swi-handler location
    gdb_debug_printf(buffer, "SWI instruction detected!\n");
    return 0x8;
}

// Note: Only the most common shift operation - left shift - is supported. Should cover most code generated by GCC.
uint32_t shift_value(uint8_t shift_field, uint32_t value, uint32_t *registers)
{
    uint32_t shift;
    if (shift_field & 1)
    {
        shift = registers[(shift_field >> 4) & 0xF] & 0xFF;
    } else
    {
        shift = (shift_field >> 3) & 0x1F;
    }
    return value << shift;
}

// Arm intermediate rotate
uint32_t rotate(uint32_t value, uint32_t shift)
{
    shift *= 2;
    return (value >> shift) | (value << (32 - shift));
}

unsigned count_bits(int n)
{     
    unsigned c; // c accumulates the total bits set in v
    for (c = 0; n; c++) 
        n &= n - 1; // clear the least significant bit set
    return c;
}

bool should_execute(uint8_t status_bits, uint8_t cond_bits)
{
    bool n = status_bits >> 3;
    bool z = (status_bits >> 2) & 1;
    bool c = (status_bits >> 1) & 1;
    bool v = status_bits & 1;

    switch (cond_bits)
    {
        case 0:
            return z;
        case 1:
            return !z;
        case 2:
            return c;
        case 3:
            return !c;
        case 4:
            return n;
        case 5:
            return !n;
        case 6:
            return v;
        case 7:
            return !v;
        case 8:
            return c && !z;
        case 9:
            return !c || z;
        case 10:
            return n == v;
        case 11:
            return n != v;
        case 12:
            return !z && (n == v);
        case 13:
            return z || (n != v);
        case 14:
            return true;
        case 15:
            return false;
        default:
            // Should not happen
            return false;
    }
}
