#include "shmac.h"
#include "serial.h"
#include "breakpoint.h"
#include "misc.h"

#define BUFFER_SIZE 1024

typedef volatile struct {
    char buffer[BUFFER_SIZE];
    unsigned head, tail, elements;
    bool char_in_gdb_packet;
} tty_state_t;

static tty_state_t state[NUM_TTY_CHANNELS];

void serial_init(unsigned channel)
{
    tty_state_t * s = &state[channel];
    s->head = s->tail = s->elements = 0;
    s->char_in_gdb_packet = false;
}

static char get_char(unsigned channel)
{
    volatile uint32_t *in = (volatile uint32_t *)((uint32_t)SYS_IN_DATA_0 | ((channel & 0xf) << 4));
    return *in;
}

static volatile uint32_t * get_output_channel(unsigned channel)
{
    return (volatile uint32_t *)((uint32_t)SYS_OUT_DATA_0 | ((channel & 0xf) << 4));
}

void serial_handle_interrupt(uint32_t * pc, uint32_t spsr, unsigned channel)
{
    // Read character from input register
    char c = get_char(channel);
    tty_state_t * s = &state[channel];    
    
    // Check if char is special interrupt package from GDB
    // Only try to trap if we are not in UND mode
    if (c == 3 && !s->char_in_gdb_packet && get_mode(spsr) != UNDEFINED)
    {
        // Insert temporary breakpoint at PC location
        breakpoint_trap(pc);
        return;
    } else if (c == '#')
    {
        s->char_in_gdb_packet = false;
    } else if (c == '$')
    {
        s->char_in_gdb_packet = true;
    }

    if (s->elements == BUFFER_SIZE)
    {
        // Drop oldest input
        serial_read_char(channel);
    }
    s->buffer[s->tail++] = c;
    if (s->tail == BUFFER_SIZE)
    {
        s->tail = 0;
    }
    ++s->elements;
}

void serial_write_char(char c, unsigned channel)
{
    volatile uint32_t *out = get_output_channel(channel);

    // wait
    while((*SYS_INT_STATUS) & (1 << channel));
    // write word
    *out = c;
}

void serial_write_string(char * s, unsigned channel)
{
    volatile uint32_t *out = get_output_channel(channel);
    for (int i = 0; s[i] != 0; i++)
    {
        // wait
        while((*SYS_INT_STATUS) & (1 << channel));
        // write word
        *out = s[i];
    }
}

char serial_read_char(unsigned channel)
{
    tty_state_t * s = &state[channel];    
    // Wait for data
    while (s->elements == 0);

    char c = s->buffer[s->head++];
    if (s->head == BUFFER_SIZE)
    {
        s->head = 0;
    }
    --s->elements;
    return c;
}
