#include "tests.h"
#include "test-misc.h"
#include "breakpoint.h"
#include "misc.h"

static uint32_t registers[16];

void test_swi(void);
void test_mov_pc(void);
void test_branch(void);
void test_add_pc(void);
void test_sub_pc(void);
void test_ldr_pc(void);
void test_ldm_pc(void);

void run_tests(void)
{
    put_string(0, "Running single step unit tests\n");
    test_swi();
    test_mov_pc();
    test_branch();
    test_add_pc();
    test_sub_pc();
    test_ldr_pc();
    test_ldm_pc();
    put_string(0, "Finished running single step unit tests\n");
}

void test_swi(void)
{
    // Name:            swi_test_1
    // Instr text:      swi 42 
    // Instr bytes:     0xef00002c
    uint32_t addr = handle_swi_instr();
    assert_equals(0x8, addr, "swi_test_1");
}

void test_mov_pc(void)
{

    // Name:            mov_pc_test_1
    // Instr text:      mov pc, lr 
    // Instr bytes:     0xe1a0f00e
    registers[R14] = 0xCAFEBABE;
    uint32_t addr1 = handle_mov_pc_instr(0xe1a0f00e, &registers[0]);
    assert_equals(registers[R14], addr1, "mov_pc_test_1");

    // Name:            mov_pc_test_2
    // Instr text:      movs pc, lr 
    // Instr bytes:     0xe1b0f00e
    registers[R14] = 0xDEADBEEF;
    uint32_t addr2 = handle_mov_pc_instr(0xe1b0f00e, &registers[0]);
    assert_equals(registers[R14], addr2, "mov_pc_test_2");

    // Name:            mov_pc_test_3
    // Instr text:      mov pc, r0 
    // Instr bytes:     0xe1a0f000
    registers[R0] = 0xBAADF00D;
    uint32_t addr3 = handle_mov_pc_instr(0xe1a0f000, &registers[0]);
    assert_equals(registers[R0], addr3, "mov_pc_test_3");
}

void test_branch(void)
{

    // Ugly hack to instantiate the two branch instructions.
    // We need to trick the compiler to believe this is not dead code.
    // Note: This test might only work with -O0 flag
    dummy_label_branch_test:
    __asm__ __volatile__ (
        "dummy_label_branch_test:\n\r"
        "b dummy_branch_end\n\r"
        "b dummy_label_branch_test\n\r"
        "bl dummy_label_branch_test\n\r"
        "dummy_branch_end:\n\r"
    );
    uint32_t dummy_label_addr = (uint32_t) &&dummy_label_branch_test;
    
    // Name:            branch_test_1
    // Instr text:      b 134
    // Instr bytes:     0xeafffffb
    uint32_t addr1 = handle_branch_instr((uint32_t * )(dummy_label_addr + 4));
    assert_equals(dummy_label_addr, addr1, "branch_test_1");

    // Name:            branch_test_2
    // Instr text:      bl 134
    // Instr bytes:     0xeafffffa
    uint32_t addr2  = handle_branch_instr((uint32_t * )(dummy_label_addr + 8));
    assert_equals(dummy_label_addr, addr2, "branch_test_2");
}

void test_add_pc(void)
{
    // Name:            add_pc_test_1
    // Instr text:      add pc, r0, #4
    // Instr bytes:     0xe280f004
    registers[R0] = 0x668;
    uint32_t addr1 = handle_add_pc_instr(0xe280f004, &registers[0]);
    uint32_t expected_addr1 = registers[R0] + 4;
    assert_equals(expected_addr1, addr1, "add_pc_test_1");
    
    // Name:            add_pc_test_2
    // Instr text:      adds pc, r0, r1
    // Instr bytes:     0xe090f001
    registers[R0] = 0x100;
    registers[R1] = 0x8;
    uint32_t addr2 = handle_add_pc_instr(0xe090f001, &registers[0]);
    uint32_t expected_addr2 = registers[R0] + registers[R1];
    assert_equals(expected_addr2, addr2, "add_pc_test_2");
    
    // Name:            add_pc_test_3
    // Instr text:      addls pc, pc, r2, lsl #2
    // Instr bytes:     0x908ff102
    registers[R2] = 0x10;
    registers[R15] = 0x1000;
    uint32_t addr3 = handle_add_pc_instr(0x908ff102, &registers[0]);
    uint32_t expected_addr3 = registers[R15] + (registers[R2] << 2) + 8;
    assert_equals(expected_addr3, addr3, "add_pc_test_3");
}

void test_sub_pc(void)
{
    // Name:            sub_pc_test_1
    // Instr text:      sub pc, lr, #4
    // Instr bytes:     0xe24ef004
    registers[R14] = 0xDEADBEEF;
    uint32_t addr1 = handle_sub_pc_instr(0xe24ef004, &registers[0]);
    uint32_t expected_addr1 = registers[R14] - 4;
    assert_equals(expected_addr1, addr1, "sub_pc_test_1"); 
    
    // Name:            sub_pc_test_2
    // Instr text:      subs pc, lr, #4
    // Instr bytes:     0xe25ef004
    registers[R14] = 0xCAFEBABE;
    uint32_t addr2 = handle_sub_pc_instr(0xe25ef004, &registers[0]);
    uint32_t expected_addr2 = registers[R14] - 4;
    assert_equals(expected_addr2, addr2, "sub_pc_test_2");

    // Name:            sub_pc_test_3
    // Instr text:      subs pc, r0, #8
    // Instr bytes:     0xe250f008
    registers[R0] = 0xBAADF00D;
    uint32_t addr3 = handle_sub_pc_instr(0xe250f008, &registers[0]);
    uint32_t expected_addr3 = registers[R0] - 8;
    assert_equals(expected_addr3, addr3, "sub_pc_test_3");    
}

void test_ldr_pc(void)
{
    uint32_t dummy_values[128];
    for (int i = 0; i < 128; i++)
    {
        dummy_values[i] = i * 42;
    }

    // Name:            ldr_pc_test_1
    // Instr text:      ldr pc, [r3, #4]
    // Instr bytes:     0xe593f004
    registers[R3] = (uint32_t) &dummy_values[0];
    uint32_t addr1 = handle_ldr_pc_instr(0xe593f004, &registers[0]);
    uint32_t expected_addr1 = ((uint32_t) &dummy_values[0]) + 4;
    assert_equals(expected_addr1, addr1, "ldr_pc_test1");

    // Name:            ldr_pc_test_2
    // Instr text:      ldr pc, [r0, r1]
    // Instr bytes:     0xe790f001
    registers[R0] = (uint32_t) &dummy_values[0];
    registers[R1] = 12;
    uint32_t addr2 = handle_ldr_pc_instr(0xe790f001, &registers[0]);
    uint32_t expected_addr2 = ((uint32_t) &dummy_values[0]) + registers[R1];
    assert_equals(expected_addr2, addr2, "ldr_pc_test2");

    // Name:            ldr_pc_test_3
    // Instr text:      ldrls pc, [pc, r3, lsl #2]
    // Instr bytes:     0x979ff103
    registers[R3] = 0x2;
    registers[R15] = (uint32_t) &dummy_values[0];
    uint32_t addr3 = handle_ldr_pc_instr(0x979ff103, &registers[0]);
    uint32_t expected_addr3 = ((uint32_t)&dummy_values[0]) + (registers[R3] << 2) + 8;
    assert_equals(expected_addr3, addr3, "ldr_pc_test3");
}

void test_ldm_pc(void)
{
    uint32_t dummy_stack[128];
    for (int i = 0; i < 128; i++)
    {
        dummy_stack[i] = i * 33;
    }
    registers[R0] = (uint32_t) &dummy_stack[32];
    registers[R13] = (uint32_t) &dummy_stack[64];

    // Name:            ldm_pc_test_1
    // Instr text:      pop {r0, r1, r2, pc}
    // Instr bytes:     0xe8bd8007
    uint32_t addr1 = handle_ldm_instr(0xe8bd8007, registers);
    uint32_t expected_addr1 = (uint32_t) &dummy_stack[64 + 3];
    assert_equals(expected_addr1, addr1, "ldm_pc_test_1");

    // Name:            ldm_pc_test_2
    // Instr text:      ldm r0, {r1, r2, pc}
    // Instr bytes:     0xe8908006
    uint32_t addr2 = handle_ldm_instr(0xe8908006, registers);
    uint32_t expected_addr2 = (uint32_t) &dummy_stack[32 + 2];
    assert_equals(expected_addr2, addr2, "ldm_pc_test_2");
    
    // Name:            ldm_pc_test_3
    // Instr text:      ldmib sp!, {r0, r1, r2, pc}
    // Instr bytes:     0xe9bd8007
    uint32_t addr3 = handle_ldm_instr(0xe9bd8007, registers);
    uint32_t expected_addr3 = (uint32_t) &dummy_stack[64 + 4];
    assert_equals(expected_addr3, addr3, "ldm_pc_test_3");
    
    // Name:            ldm_pc_test_4
    // Instr text:      ldmdb sp!, {r0, r1, r2, pc}
    // Instr bytes:     0xe93d8007
    uint32_t addr4 = handle_ldm_instr(0xe93d8007, registers);
    uint32_t expected_addr4 = (uint32_t) &dummy_stack[64 - 1];
    assert_equals(expected_addr4, addr4, "ldm_pc_test_4");
    
    // Name:            ldm_pc_test_5
    // Instr text:      ldmda sp!, {r0, r1, r2, pc}
    // Instr bytes:     0xe83d8007
    uint32_t addr5 = handle_ldm_instr(0xe83d8007, registers);
    uint32_t expected_addr5 = (uint32_t) &dummy_stack[64];
    assert_equals(expected_addr5, addr5, "ldm_pc_test_5");

    // Name:            ldm_pc_test_6
    // Instr text:      pop {lr, pc}
    // Instr bytes:     0xe8bdc000
    uint32_t addr6 = handle_ldm_instr(0xe8bdc000, registers);
    uint32_t expected_addr6 = (uint32_t) &dummy_stack[64 + 1];
    assert_equals(expected_addr6, addr6, "ldm_pc_test_6");

}