#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <stdint.h>

#include "shmac.h"
#include "gdb.h"
#include "mutex.h"

static volatile unsigned interrupt_count;
static volatile bool swi_happened;
static volatile bool undef_happened;

static volatile uint32_t print_mutex;

static void putChar(int channel, uint8_t c) {
    volatile uint32_t *sys_out_data = (volatile uint32_t *)((uint32_t)SYS_OUT_DATA_0 | ((channel & 0xf) << 4));

    // wait
    while((*SYS_INT_STATUS) & (1 << channel));
    // write word
    *sys_out_data = c;
}


static uint8_t getChar(int channel) {
  volatile uint32_t *sys_in_data =
    (volatile uint32_t *)((uint32_t)SYS_IN_DATA_0 | ((channel & 0xf) << 4));

  // read word
  uint8_t c = *sys_in_data;
  return c;
}


static void putUint4(int channel, uint8_t val) {
  if(val > 9) putChar(channel, val - 10 + 'A');
  else putChar(channel, val + '0');
}

static void putUint8(int channel, uint8_t val) {
  putUint4(channel, (val >> 4) & 0xf);
  putUint4(channel, val & 0xf);
}

static void putString(int channel, char *s) {
    while(*s != 0) {
        if(*s == '\n') putChar(channel, '\r');
        putChar(channel, *s);
        s++;
    }
}

static void user_print_str(char * str)
{
    lock_mutex(&print_mutex);
    // int channel = *TILEREG_CPUID;
    int channel = 0;
    putString(channel, str);
    unlock_mutex(&print_mutex);
}

int dummy(int value)
{
    return 42+value;
}

void undef_instr_c_handler(void)
{
    user_print_str("C undef handler called\n");
}

void init_timed_interrupts(void)
{
    // enable host and timer0 interrupts
    *IC0_IRQ_ENABLESET |= INT_MASK_TIMER0;

    // enable timer 0
    *TIMER0_LOAD = 60000; // 16 bit register, with implicit left shift 8
    *TIMER0_CTRL = TIMER_CTRL_ENABLE | TIMER_CTRL_PERIODIC | TIMER_CTRL_SCALE_1;
}

void swi_c_handler(uint32_t lr)
{
    user_print_str("C swi handler called\n");
}

void irq_c_handler(void)
{
    uint32_t irqStatus = *IC0_IRQ_STATUS;
    int channel;
    int ints = irqStatus >> 16;

    for(channel = 0; channel < NUM_TTY_CHANNELS; channel++)
    {
        if(ints & (1 << channel))
        {
            // read host char and clear host interrupt
            uint32_t x = getChar(channel);
            putChar(0, '0' + channel);
            putChar(0, ':');
            putChar(0, x);
            putChar(0, '\r');
            putChar(0, '\n');
        }
    }

    if(irqStatus & INT_MASK_TIMER0)
    {
        putChar(0, '.');
        // clear timer interrupt
        *TIMER0_CLR = 1;
    }
}

void die(void)
{
    user_print_str("killed\r\n");
}

int main(void)
{    
    int cpu_id = *TILEREG_CPUID;
    if (cpu_id == 0)
    {
        init_mutex(&print_mutex);
        gdb_config_t gdb_config;
        gdb_config.tty_channel[0] = 1;
        gdb_config.tty_channel[1] = 2;
        gdb_config.tty_channel[2] = 3;
        gdb_config.kill_program_handler = &die;
        gdb_init(&gdb_config);
        *SYS_READY = 1;
    }
    GDB_TRAP;
    init_timed_interrupts();

    int i = 0;
    int x = dummy(8) % 9;
    char my_stack_buffer[100];

    if (my_stack_buffer[45] == 0)
    {   
        my_stack_buffer[0] = '0';
    } else
    {
        my_stack_buffer[0] = '1';
    }
    for (int j = 0; j < 10; j++)
    {
        i += 10;
        if (j % 4 == 0)
        {
            i = j + i;
        }
        if (i > 40)
        {
            i += 10;
        }
    }
    __asm__ ("mcr p1, 6, R13, c9, c5, 1");
    __asm__ ("swi 42");
    switch (x)
    {
        case 0:
            i *= 2;
            break;
        case 1:
            i += 3;
            break;
        case 2:
            i -= 3;
            break;
        case 3:
            i *= 3;
            break;
        case 4:
            i -= 93;
            break;
        case 5:
            i -= 13;
            break;
        case 6:
            i -= 43;
            break;
        case 7:
            i -= 32;
            break;
        case 8:
            i -= 13;
            break;
    }
    bool forevah = true;
    while (forevah)
    {
        i = i * 10 - 3;
    }

    return i;
}

// This part is needed for debug prints in breakpoints to work

char prog_top;    /* Defined by the linker.  */
#define ENOMEM          12      /* Out of memory */
#include <errno.h>

caddr_t _sbrk(int incr)
{
    register char *stack_ptr __asm__("sp");

    static char *heap_end = NULL;
    char *prev_heap_end = NULL;

    if (heap_end == NULL)
        heap_end = &prog_top;

    prev_heap_end = heap_end;

    //printf("0x%08x + %d >= 0x%08x ? ", heap_end, incr, stack_ptr);
    if (heap_end + incr >= stack_ptr) {
        //printf("yes\r\n");
        errno = ENOMEM;
        return (caddr_t) - 1;
    }
    //printf("no\r\n");

    heap_end += incr;

    return (caddr_t) (incr > 0 ? prev_heap_end : heap_end);
}

