#include "gdb/enable.h"	
.global gdb_irq_handler
.global gdb_undef_instr_handler

#include "gdb/defs.h"

#ifndef DISABLE_DEBUGGER

gdb_irq_handler:
    // Save regs
    push {r0-r3,r9,r12,lr}
    
    // PC is the first argument
    sub r0, lr, #4

    // SPSR is second argument
    mrs r1, spsr

    // Jump to C function
    bl gdb_c_irq_handler

    // Check if exception should be dispatched to user handler
    cmp r0, #1
    beq $dispatch_irq
    
    // Restore regs
    pop {r0-r3,r9,r12,lr}

    // Jump to user program's irq handler
    subs pc, lr, #4

// Dispatch to user program's handler
$dispatch_irq:
    // Restore regs
    pop {r0-r3,r9,r12,lr}

    // Branch to user handler
    b gdb_irq_handler_program

gdb_undef_instr_handler:
    // Make room for 18 registers on stack
    sub sp, sp, #72
    // Store the first 13 registers to stack
    stmia sp!, {r0-r12}
    // lr = instruction after where to return
    sub r0, lr, #4
    // Store PC
    str r0, [sp, #8]

    // Read spsr (previous mode's cpsr)
    mrs r1, spsr
    // Store it as previous mode's cpsr
    str r1, [sp, #12]
    // Filter out the irrelevant bits
    and r2, r1, #0xF

    // Store banked registers (sp + lr + spsr)
    cmp r2, #0
    bleq $store_user_banked_registers
    cmp r2, #1
    bleq $store_fiq_banked_registers
    cmp r2, #2
    bleq $store_irq_banked_registers
    cmp r2, #3
    bleq $store_svc_banked_registers
    cmp r2, #7
    bleq $store_abort_banked_registers
    cmp r2, #11
    bleq $store_undef_banked_registers

    // pass pointer to registers as argument to handler
    sub sp, sp, #72
    mov r0, sp

    // Jump to gdb stub handler
    bl gdb_exception_handler

    // Load previous sp
    ldr r1, [sp, #52]
    // Load previous lr
    ldr r2, [sp, #56]
    // Load stored cpsr and set it as cpsr
    ldr r3, [sp, #64]
    msr spsr, r3
    // Get mode bits from cpsr
    and r4, r3, #0xF
    // Branch to load correct banked registers
    cmp r4, #0
    bleq $restore_user_banked_registers
    cmp r4, #1
    bleq $restore_fiq_banked_registers
    cmp r4, #2
    bleq $restore_irq_banked_registers
    cmp r4, #3
    bleq $restore_svc_banked_registers
    cmp r2, #7
    bleq $restore_abort_banked_registers

    // Point r1 to save area
    add r1, sp, #4
    // Restore most registers
    ldmia r1, {r1-r12}
    // Fetch return address to lr (previous pc)
    ldr lr, [sp, #60]
    // Check if exception should be dispatched to user handler
    cmp r0, #1
    beq $dispatch_exception 
    // Restore r0
    ldr r0, [sp]
    // Restore sp
    add sp, sp, #72
    // Return
    movs pc, lr

// Dispatch to user program's handler
$dispatch_exception:
    // Restore r0
    ldr r0, [sp]
    // Restore sp
    add sp, sp, #72
    // Make lr point to next instruction
    add lr, lr, #4
    // Return
    b gdb_undef_instr_handler_program

$store_user_banked_registers:
    // Store user sp and lr
    stmia sp, {sp, lr}^
    nop
    nop
    // Increment sp to spsr position
    add sp, sp, #16
    // Store 0 as spsr
    mov r0, #0
    str r0, [sp]!
    add sp, sp, #4
    // Return
    mov pc, lr

$store_fiq_banked_registers:
    mov r0, #0xD1
    b $store_privileged_registers
    
$store_irq_banked_registers:
    mov r0, #0xD2
    b $store_privileged_registers

$store_svc_banked_registers:
    mov r0, #0xD3
    b $store_privileged_registers

$store_abort_banked_registers:
    mov r0, #0xD7
    b $store_privileged_registers

$store_undef_banked_registers:
    // Calculate original sp
    add r0, sp, #20
    // Store stack and lr
    stmia sp!, {r0, lr}
    // Get spsr register
    mrs r0, spsr
    str r0, [sp, #8]!
    add sp, sp, #4
    // Return
    mov pc, lr

$store_privileged_registers:
    // Backup current cpsr
    mrs r1, cpsr
    // Change mode
    msr cpsr, r0
    nop
    // Copy banked registers
    mov r2, sp
    mov r3, lr
    mrs r4, spsr
    // Change back to UND
    msr cpsr, r1
    nop
    // Store sp and lr to stack
    stmia sp!, {r2, r3}
    str r4, [sp, #8]!
    add sp, sp, #4
    // Return
    mov pc, lr

$restore_user_banked_registers:
    // Store user sp and lr
    push {r1, r2}
    ldm sp, {sp, lr}^
    nop
    nop
    // Decrement stack
    add sp, sp, #8
    mov pc, lr

$restore_fiq_banked_registers:
    mov r5, #0xD1
    b $restore_privileged_registers
    
$restore_irq_banked_registers:
    mov r5, #0xD2
    b $restore_privileged_registers

$restore_svc_banked_registers:
    mov r5, #0xD3
    b $restore_privileged_registers

$restore_abort_banked_registers:
    mov r5, #0xD4
    b $restore_privileged_registers

$restore_privileged_registers:
    // Load stored spsr
    ldr r7, [sp, #68]
    // Backup current cpsr
    mrs r6, cpsr
    // Change mode
    msr cpsr, r5
    nop
    // Restore sp and lr
    mov sp, r1
    mov lr, r2
    // Restore spsr
    msr spsr, r7
    // Change back to UND mode
    msr cpsr, r6
    nop
    mov pc, lr

#else

gdb_irq_handler:
    b gdb_irq_handler_program

gdb_undef_instr_handler:
    b gdb_undef_instr_handler_program
#endif
