#include "gdb/breakpoint.h"
#include "gdb/enable.h"
#include "gdb/gdb-stub.h"
#include "gdb/misc.h"
#include "gdb/packet.h"
#include "gdb/process.h"
#include "gdb/serial.h"
#include "gdb/defs.h"
#include "offsets.h"
#include <barrelfish/gdb_syscalls.h>

#define BUFFER_SIZE 4096
typedef struct {
    char input_buffer[BUFFER_SIZE];
    char output_buffer[BUFFER_SIZE];
    bool is_range_stepping;
    uint32_t range_step_start;
    uint32_t range_step_end;
    bool debug_enabled;
    unsigned tty_channel;
    uint32_t ic0_enable_set;
    uint32_t ic1_enable_set;
} core_state_t;

static core_state_t state[NUM_CORES];
static unsigned tty_reserved_bitmap;
static void (*kill_program_handler) (void);

const char * SUPPORTED_FEATURES_STRING = 
    "multiprocess+;BreakpointCommands+;qRelocInsn-;ConditionalBreakpoints-;ConditionalTracepoints-;qXfer:libraries:read+";

static uint8_t get_signal(uint32_t *registers, operating_mode_t mode);
static bool prepare_return(uint8_t gdb_signal, core_state_t * s, uint32_t * registers);
static void send_reply_packet(uint8_t gdb_signal, uint32_t *registers, core_state_t * s);
static void enable_host_interrupt(core_state_t * s);
static bool is_gdb_syscall(uint32_t * registers);
static void handle_gdb_syscall(uint32_t * registers);

void gdb_init(gdb_config_t * config)
{
#ifndef DISABLE_DEBUGGER
    breakpoint_init();
    gdb_process_init();

    tty_reserved_bitmap = 0;
    kill_program_handler = config->kill_program_handler;

    for (unsigned i = 0; i < NUM_CORES; i++)
    {
        core_state_t * s = &state[i];
        int tty_channel = config->tty_channel[i];
        if (tty_channel < 0)
        {
            s->debug_enabled = false; 
        } else {
            s->debug_enabled = true;
            s->tty_channel = tty_channel;
            tty_reserved_bitmap |= (1 << tty_channel);

            gdb_serial_init(tty_channel);
        }
        s->is_range_stepping = false;
    }
#endif 
}

void gdb_print(char * string)
{
    unsigned id = get_cpu_id();
    char *out = state[id].output_buffer;
    unsigned channel = state[id].tty_channel;
    *out++ = 'O';
    int length = string_length(string);
    mem_to_hex(string, out, length);
    put_packet(state[id].output_buffer, channel);
}

// Handles IO
bool gdb_c_irq_handler(uint32_t * pc, uint32_t spsr)
{    
    uint32_t irq_status = *IC0_IRQ_STATUS;
    
    unsigned channel_bitmap = irq_status >> 16;

    for (unsigned channel = 0; channel < NUM_TTY_CHANNELS; channel++)
    {
        if (channel_bitmap & tty_reserved_bitmap & (1 << channel))
        {
            serial_handle_interrupt(pc, spsr, channel);
        }
    }
    // We do only forward interrupts to program if all GDB ttys are consumed
    return (channel_bitmap & tty_reserved_bitmap) == 0;
}

bool gdb_exception_handler(uint32_t * registers)
{
    unsigned id = get_cpu_id();
    core_state_t * s = &state[id];
    operating_mode_t mode = get_mode(registers[CPSR]);
    uint8_t gdb_signal = get_signal(registers, mode);
    uint32_t pid = get_current_pid();
    unsigned channel = s->tty_channel;

    if (is_gdb_syscall(registers))
    {
        handle_gdb_syscall(registers);
        return prepare_return(gdb_signal, s, registers);
    }

    // Return to CPUs handler if debugging is not enabled for this core
    if (!s->debug_enabled)
    {
        return prepare_return(gdb_signal, s, registers);
    }

    // Clear any trap breakpoint
    breakpoint_clear_trap();

    enable_host_interrupt(s);

    if (mode == UNDEFINED)
    {
        gdb_print("FATAL ERROR: The program hit a breakpoint while executing in 'undefined' mode.\n"
            "You will not be able to continue executing the program after returning from GDB (lr + spsr are lost).\n"
            "Make sure to not debug code shared by the GDB stub or your undefined instruction handler.\n");
        put_packet("S09", channel);
    } else if (s->is_range_stepping)
    {
        uint32_t pc = registers[R15];
        // Check if we are in range for range step
        if (pc >= s->range_step_start && pc < s->range_step_end)
        {
            // Insert single step breakpoint
            breakpoint_single_step(registers, (uint32_t *) pc);
            // Return back to program debugged
            return prepare_return(gdb_signal, s, registers); 
        } else
        {
            s->is_range_stepping = false;
            send_reply_packet(gdb_signal, registers, s);
        }
    } else
    {
        send_reply_packet(gdb_signal, registers, s);
    }

    while (1)
    {
        char *packet = read_packet(s->input_buffer, BUFFER_SIZE, channel);
        char cmd = *packet++;
        switch (cmd)
        {
            // Enable extended mode - the stub assumes that this mode is always enabled
            case '!':
            {
                put_packet("OK", channel);
                break;
            }

            // Get last signal
            case '?':
            {
                send_reply_packet(gdb_signal, registers, s);
                break;
            }

            case 'D':
            {
                // TODO: Detach process
                put_packet("OK", channel);
                break;
            }

            // Set thread for sub-sequent operations - we ignore this as we are only using a single thread per process
            case 'H':
            {
                put_packet("OK", channel);
                break;
            }

            // Terminate system
            case 'k':
            {
                (*kill_program_handler)(); 
                break;
            }

           // Read all standard registers
            case 'g':
            {
                // Copy register R0 - R15 (note that CPSR is not included)
                mem_to_hex((char *) registers, s->output_buffer, 64);
                put_packet(s->output_buffer, channel);
                break;
            }

            // Write to all standard registers
            case 'G':
            {
                // Copy register R0 - R15 (note that CPSR is not included)
                hex_to_mem(packet, (char *) registers, 64);
                put_packet("OK", channel);
                break;
            }

            // Read memory from given location of specified length
            case 'm':
            {
                uint32_t address, length;
                if (hex_to_int(&packet, &address) && 
                    *packet++ == ',' && 
                    hex_to_int(&packet, &length))
                {
                    mem_to_hex((char*) address, s->output_buffer, length);
                    put_packet(s->output_buffer, channel);
                } else 
                {
                    put_packet("E01", channel);
                }
                break;
            }

            // Write memory from given location of specified length
            case 'M':
            {
                uint32_t address, length;
                if (hex_to_int(&packet, &address) && 
                    *packet++ == ',' && 
                    hex_to_int(&packet, &length) &&
                    *packet++ == ':')
                {
                    hex_to_mem(packet, (char*) address, length);    
                    put_packet("OK", channel);
                } else
                {
                    put_packet("E01", channel);
                }
                break;
            }

            // Write binary data to given location
            case 'X':
            {
                uint32_t address, length;
                if (hex_to_int(&packet, &address) && 
                    *packet++ == ',' && 
                    hex_to_int(&packet, &length) &&
                    *packet++ == ':')
                {
                    uint8_t *source = (uint8_t*) packet;
                    uint8_t *dest = (uint8_t*) address;
                    mem_copy(dest, source, length);
                    put_packet("OK", channel);
                } else
                {
                    put_packet("E01", channel);
                }
                break;                
            }

            // Continue
            case 'c':
            {
                uint32_t address;
                if (hex_to_int(&packet, &address))
                {
                    registers[R15] = address;
                }
                // Return back to program debugged
                return prepare_return(gdb_signal, s, registers);
            }

            // Read specific register
            case 'p':
            {
                uint32_t register_number;
                if (hex_to_int(&packet, &register_number))
                {
                    if (register_number == GDB_CPSR_ID)
                    {
                        uint32_t cpsr = registers[CPSR];
                        four_bytes_to_hex(s->output_buffer, cpsr);
                        put_packet(s->output_buffer, channel);
                    } else if (register_number == GDB_SPSR_ID)
                    {
                        uint32_t spsr = registers[SPSR];
                        four_bytes_to_hex(s->output_buffer, spsr);
                        put_packet(s->output_buffer, channel); 
                    } else if (register_number < 16)
                    {
                        four_bytes_to_hex(s->output_buffer, registers[register_number]);
                        put_packet(s->output_buffer, channel);

                    } else {
                        // xx for unknown registers
                        put_packet("xx", channel);
                    }
                } else
                {
                    put_packet("E01", channel);
                }
                break;
            }

            // Write specific register
            case 'P':
            {
                uint32_t register_number;
                if (hex_to_int(&packet, &register_number) &&
                    *packet++ == '=')
                {
                    if (register_number < 16)
                    {
                        hex_to_mem(packet, (char *)&registers[register_number], 4);
                        put_packet("OK", channel);
                    } else if (register_number == GDB_SPSR_ID)
                    {
                        hex_to_mem(packet, (char *)&registers[SPSR], 4);
                        put_packet("OK", channel); 
                    } else if (register_number == GDB_CPSR_ID)
                    {
                        hex_to_mem(packet, (char *)&registers[CPSR], 4);
                        put_packet("OK", channel);
                    } else {
                        put_packet("E01", channel);
                    }
                } else {
                    put_packet("E02", channel);
                }
                break;
            }

            // Generic case for all query packages (all starts with letter 'q')
            case 'q':
            {
                // qC - return current thread-id / pid
                if (matches(packet, "C"))
                {
                    char * out = s->output_buffer;
                    out = string_copy(out, "QCp");
                    out = int_to_string(out, pid);
                    out = string_copy(out, ".-1");
                    put_packet(s->output_buffer, channel);
                // qSupported - query stub for supported features
                } else if (matches(packet, "Supported"))
                {
                    put_packet(SUPPORTED_FEATURES_STRING, channel);
                // qOffset - return offset of program sections
                } else if (matches(packet, "Offsets"))
                {
                    write_core_reloc(s->output_buffer);
                    put_packet(s->output_buffer, channel);
                // qAttached - tells stub if process was created or attached to. We are always attaching for kernel processes
                } else if (matches(packet, "Attached"))
                {
                    put_packet("1", channel);
                // qRcmd - custom command register
                } else if (matches(packet, "Rcmd,"))
                {
                    packet += 5;
                    unsigned length = string_length(packet);
                    // Set max cap for command to be 64 chars
                    if (length > 64) length = 64;
                    // Hex encoding takes twice as much storage as raw data 
                    char command_str[33] = {0};
                    hex_to_mem(packet, command_str, length);

                    if (matches(command_str, "processes"))
                    {
                        print_processes();
                        put_packet("OK", channel);
                    } else 
                    {
                        gdb_print("Unknown command '");
                        gdb_print(command_str);
                        gdb_print("'.\n");
                        gdb_print(
                            "Available commands:\n"
                            " * processes - list all processes.\n");
                        put_packet("OK", channel);
                    }
                // qXfer:libraries:read
                } else if (matches(packet, "Xfer:libraries:read::"))
                {
                    packet += 21;
                    uint32_t offset, length;
                    if (hex_to_int(&packet, &offset) && 
                        *packet++ == ',' && 
                        hex_to_int(&packet, &length))
                    {
                        write_library_list(s->output_buffer, offset, length);
                        put_packet(s->output_buffer, channel);
                    } else
                    {
                        put_packet("E01", channel);
                    }                
                } else
                {
                    *s->output_buffer = 0;
                    put_packet(s->output_buffer, channel);
                }
                break;
            }

            // Generic case for "verbose" packages. Typically newer additions to the protocol
            case 'v':
            {
                if (matches(packet, "Attach;"))
                {
                    packet += 7;
                    uint32_t pid;
                    if (hex_to_int(&packet, &pid))
                    {
                        attach_process(pid);
                        return prepare_return(gdb_signal, s, registers);
                    } else
                    {
                        put_packet("E01", channel);
                    }
                // vKill - kill process attached
                } else if (matches(packet, "Kill"))
                {
                    //TODO: Kill process
                    put_packet("OK", channel);
                // vCont? - query package for all supported actions
                } else if (matches(packet, "Cont?"))
                {
                    // Note: GDB require us to support C and S (continue and step with signal).
                    // We will make their implementation ignore the signal
                    put_packet("vCont;c;C;s;S;r", channel);
                } else
                // vCont - general command for control flow commands (step, continue, range step)
                if (matches(packet, "Cont"))
                {
                    packet += 5;
                    uint32_t *pc = (uint32_t *) registers[R15];
                    switch (*packet++)
                    {
                        case 's':
                        case 'S':
                        {
                            // Insert single step breakpoint
                            breakpoint_single_step(registers, pc);
                            // Return back to program debugged
                            return prepare_return(gdb_signal, s, registers); 
                        }
                        case 'c':
                        case 'C':
                        {
                            // Return back to program debugged
                            return prepare_return(gdb_signal, s, registers);
                        }
                        case 'r':
                        {
                            uint32_t start, end;
                            if (hex_to_int(&packet, &start) &&
                                *packet++ == ',' &&
                                hex_to_int(&packet, &end))
                            {
                                s->is_range_stepping = true;
                                s->range_step_start = start;
                                s->range_step_end = end;
                                // Insert single step breakpoint
                                breakpoint_single_step(registers, pc);
                                // Return back to program debugged
                                return prepare_return(gdb_signal, s, registers); 

                            } else
                            {
                                put_packet("E01", channel);
                            }
                            break;
                        }
                        default:
                        {
                            put_packet("E02", channel);
                        }
                    }
                } else
                {
                     *s->output_buffer = 0;
                    put_packet(s->output_buffer, channel);
                }
                break;
            }

            // Single step instruction
            case 's':
            {
                uint32_t resume_addr;
                if (hex_to_int(&packet, &resume_addr))
                {
                    // Change PC to resume_addr
                    registers[R15] = resume_addr;
                } else
                {
                    // We should resume from PC
                    resume_addr = registers[R15];
                }   

                // Poke into memory and get the instruction
                uint32_t *resume_instr = (uint32_t *) resume_addr;
                // Insert single step breakpoints 
                breakpoint_single_step(registers, resume_instr);
                // Return back to program debugged
                return prepare_return(gdb_signal, s, registers);                
            }

            // Insert breakpoint
            case 'Z':
            {
                char type = *packet++;
                uint32_t address;
                uint32_t kind;
                if (type != '0')
                {
                    *s->output_buffer = 0;
                    put_packet(s->output_buffer, channel);
                } else if (*packet++ == ',' &&
                    hex_to_int(&packet, &address) && 
                    *packet++ == ',' && 
                    hex_to_int(&packet, &kind))
                {
                    if (kind != 4)
                    {
                        *s->output_buffer = 0;
                        put_packet(s->output_buffer, channel);
                    } else 
                    {
                        uint32_t *instr = (uint32_t *) address;
                        if (breakpoint_insert(instr))
                        {
                            put_packet("OK", channel);
                        } else
                        {
                            put_packet("E02", channel);
                        }
                    }
                } else
                {
                    put_packet("E01", channel);
                }
                break;
            }

            // Remove breakpoint
            case 'z':
            {
                char type = *packet++;
                uint32_t address;
                uint32_t kind;
                if (type != '0')
                {
                    *s->output_buffer = 0;
                    put_packet(s->output_buffer, channel);
                } else if (*packet++ == ',' &&
                    hex_to_int(&packet, &address) && 
                    *packet++ == ',' && 
                    hex_to_int(&packet, &kind))
                {
                    if (kind != 4)
                    {
                        *s->output_buffer = 0;
                        put_packet(s->output_buffer, channel);
                    } else
                    {
                        uint32_t *instr = (uint32_t *) address;
                        breakpoint_delete(instr);
                        put_packet("OK", channel);
                    }
                } else
                {
                    put_packet("E01", channel);
                }
                break;
            }

            // Unknown command
            default:
            {
                *s->output_buffer = 0;
                put_packet(s->output_buffer, channel);
                break;
            }
                
        }
    }
}


static uint8_t get_signal(uint32_t * registers, operating_mode_t mode)
{
    if (mode == ABORT)
    {
        return SIGBUS;
    }
    // Access instruction at PC location
    uint32_t * pc = (uint32_t *) registers[R15];
    if (*pc == STANDARD_BREAKPOINT_INSTR || *pc == STATIC_BREAKPOINT_INSTR || 
        *pc == STEP_BREAKPOINT_INSTR)
    {
        return SIGTRAP;
    } else if (*pc == GDB_SYSCALL_INTERRUPT_INSTR || *pc == INTERRUPT_BREAKPOINT_INSTR)
    {
        return SIGINT;
    }
    return SIGILL;
}

static bool prepare_return(uint8_t gdb_signal, core_state_t * s, uint32_t * registers)
{
    // Disable interrupts
   __asm__ __volatile__ (
        "push {r0}\n\t"
        "mrs r0, cpsr\n\t"
        "orr r0, r0, #0x80\n\t"
        "msr cpsr, r0\n\t"
        "pop {r0}\n\t"
    );

    // Restore interrupt mask
    // Note: Host interrupt for GDB is kept enabled
    *IC0_IRQ_ENABLESET = s->ic0_enable_set;
    *IC1_IRQ_ENABLESET = s->ic1_enable_set;

    // Flush instruction cache
    __asm__ __volatile__ (
        "push {r0}\n\t"
        "mov r0, #0\n\t"
        "mcr p15, 0, r0, c7, c5, 0\n\t"
        "pop {r0}\n\t"
    );

    // Jump over static/compiled breakpoints
    uint32_t * pc = (uint32_t *) registers[R15];
    if (*pc == STATIC_BREAKPOINT_INSTR || *pc == GDB_SYSCALL_INTERRUPT_INSTR)
    {
        registers[R15] += 4;
    }

    // Dispatch exception to user handler only if the invalid instruction was not a recognized breakpoint
    return gdb_signal == SIGILL;
}

static void send_reply_packet(uint8_t gdb_signal, uint32_t *registers, core_state_t * s)
{
    char *out = s->output_buffer;
    *out++ = 'T';
    out = byte_to_hex(out, gdb_signal);
    for (uint8_t reg = R11; reg <= R15; reg++)
    {
        out = byte_to_hex(out, reg);
        *out++ = ':';
        out = four_bytes_to_hex(out, registers[reg]);
        *out++ = ';';
    }
    out = byte_to_hex(out, GDB_CPSR_ID);
    *out++ = ':';
    out = four_bytes_to_hex(out, registers[CPSR]);
    *out++ = ';';
    out = string_copy(out, "core:");
    out = int_to_hex(out, get_cpu_id());
    *out++ = ';';
    if (is_library_xml_updated())
    {
        out = string_copy(out, "library:nil;");
    }

    *out = 0;
    put_packet(s->output_buffer, s->tty_channel);
}

static void enable_host_interrupt(core_state_t * s)
{
    s->ic0_enable_set = *IC0_IRQ_ENABLESET;
    s->ic1_enable_set = *IC1_IRQ_ENABLESET;

    // Disable all interrupts
    *IC0_IRQ_ENABLECLR = 0xFFFF001F;
    *IC1_IRQ_ENABLECLR = 0xFFFF001F;

    // Enable host interrupts for tty channel
    *IC0_IRQ_ENABLESET = (1 << (s->tty_channel + 16));

    // Enable interrupts for CPU
    __asm__ __volatile__ (
        "push {r0}\n\t"
        "mrs r0, cpsr\n\t"
        "bic r0, r0, #0x80\n\t"
        "msr cpsr, r0\n\t"
        "pop {r0}\n\t"
    );
}

static bool is_gdb_syscall(uint32_t * registers)
{
    uint32_t * pc = (uint32_t *) registers[R15];
    return *pc == GDB_SYSCALL_INTERRUPT_INSTR; 
}

static void handle_gdb_syscall(uint32_t * registers)
{
    syscall_ident_t type = (syscall_ident_t) registers[R0];
    switch (type)
    {
        case PROCESS_CREATED:
        {
            process_created_params_t * params = (process_created_params_t *) registers[R1];
            gdb_on_process_created(params->handle, params->segment_address);
            break;
        }

        case PROCESS_DESTROYED:
        {
            process_destroyed_params_t * params = (process_destroyed_params_t *) registers[R1];
            gdb_on_process_destroyed(params->handle);
            break;
        }

    }
}
