#include "gdb/enable.h"
#include "gdb/misc.h"
#include "gdb/serial.h"
#include "offsets.h"

static const char hexchars[] = "0123456789abcdef";
static const char digits[] = "0123456789";

const uint32_t GDB_SYSCALL_INTERRUPT_INSTR = 0xeec9d1b4; // mcr p1, 6, R13, c9, c4, 5
const uint32_t STEP_BREAKPOINT_INSTR = 0xeec9d194; // mcr p1, 6, R13, c9, c4, 4
const uint32_t STANDARD_BREAKPOINT_INSTR = 0xeec9d174; // mcr p1, 6, R13, c9, c4, 3
const uint32_t INTERRUPT_BREAKPOINT_INSTR = 0xeec9d154; // mcr p1, 6, R13, c9, c4, 2
const uint32_t STATIC_BREAKPOINT_INSTR = 0xeec9d134; // mcr p1, 6, R13, c9, c4, 1

static uint8_t hex(char ch)
{
    if (ch >= 'a' && ch <= 'f')
        return ch-'a'+10;
    if (ch >= '0' && ch <= '9')
        return ch-'0';
    if (ch >= 'A' && ch <= 'F')
        return ch-'A'+10;
    return 0;
}

int hex_to_int(char **ptr, uint32_t *intValue)
{
    int numChars = 0;
    *intValue = 0;
    while (**ptr != ',' && **ptr != ':' && **ptr != '=' && **ptr != 0)
    {
        int hexValue = hex(**ptr);
        *intValue = (*intValue << 4) | hexValue;
        numChars++;

        (*ptr)++;
    }
    return numChars;
}

uint8_t hex_to_byte(char ch1, char ch2)
{
    return (hex(ch1) << 4) + hex(ch2);
}

char * byte_to_hex(char *ptr, uint8_t c)
{
    *ptr++ = hexchars[c >> 4];
    *ptr++ = hexchars[c & 0xf];
    *ptr = 0;
    return ptr;
}

char * four_bytes_to_hex(char *ptr, uint32_t c)
{
    return mem_to_hex((char *)&c, ptr, 4);
}

char * int_to_hex(char *ptr, uint32_t c)
{
    ptr = byte_to_hex(ptr, (c >> 24) & 0xFF);
    ptr = byte_to_hex(ptr, (c >> 16) & 0xFF);
    ptr = byte_to_hex(ptr, (c >>  8) & 0xFF);
    ptr = byte_to_hex(ptr, (c >>  0) & 0xFF);
    return ptr;
}


char * mem_to_hex(char *mem, char *buf, int count)
{
    while (count-- > 0)
    {
        char ch = *mem++;
        *buf++ = hexchars[ch >> 4];
        *buf++ = hexchars[ch & 0xf];
    }
    *buf = 0;
    return buf;
}

char * hex_to_mem(char *buf, char *mem, int count)
{
    for (int i = 0; i < count; ++i)
    {
        char ch = hex(*buf++) << 4;
        ch |= hex(*buf++);
        *mem++ = ch;

    }
    return mem;
}

char * string_copy(char *dest, const char *source)
{
    while ((*dest++ = *source++));
    return dest - 1;
}

void mem_copy(uint8_t *dest, const uint8_t *source, uint32_t length)
{
    for (unsigned int i = 0; i < length; i++)
    {
        dest[i] = source[i];
    }
}

int string_length(const char *str)
{
    const char *s;
    for (s = str; *s; ++s);
    return (s - str);
}

bool matches(const char *packet, const char *pattern)
{
    for (int i = 0; pattern[i] != 0; ++i)
    {
        if (packet[i] != pattern[i])
        {
            return false;
        }
    }
    return true;
}

char * int_to_string(char *ptr, uint32_t value)
{
    int shifter = value;
    do
    {
        ++ptr;
        shifter /= 10;
    } while(shifter);
    *ptr = '\0';
    char * end = ptr;
    do
    {
        *--ptr = digits[value % 10];
        value /= 10;
    } while(value);
    return end;
}

operating_mode_t get_mode(uint32_t spsr)
{
    switch (spsr & 0xF)
    {
        case 0:
            return USER;

        case 1:
            return FIQ;
        
        case 2:
            return IRQ;

        case 3:
            return SVC;

        case 7:
            return ABORT;

        case 11:
            return UNDEFINED;

        // Should not happen
        default:
            return UNDEFINED;
    }
}

unsigned get_cpu_id(void)
{
    return *TILEREG_CPUID;
}