#include "gdb/breakpoint.h"
#include "gdb/enable.h"
#include "gdb/gdb-stub.h"
#include "gdb/misc.h"
#include "gdb/process.h"
#include "gdb/defs.h"
#include "gdb/mutex.h"
#include <dispatch.h>

#include <kernel.h>
#include <barrelfish/dispatcher_arch.h>

static volatile uint32_t mutex;

typedef volatile struct {
    dispatcher_handle_t handle;
    uint32_t pid;
    uint32_t segment_address;
    char name[32];
    unsigned core_id;
} process_t;

typedef volatile struct {
    uint32_t attached_pid;
    bool is_waiting_for_process;
    bool library_xml_updated;
} core_state_t;


static volatile process_t processes[MAX_PROCESS_ENTRIES];
static volatile char library_xml[2048];
static volatile uint32_t pid_generator;
static volatile core_state_t core_state[NUM_CORES];
static volatile uint32_t core_text_reloc[NUM_CORES];

static dispatcher_handle_t find_dispatcher(uint32_t pid);
static void trap_dispatcher(dispatcher_handle_t handle);
static void update_library_xml(void);

void gdb_process_init(void)
{
    gdb_init_mutex(&mutex);
    for (unsigned i = 0; i < MAX_PROCESS_ENTRIES; i++)
    {
        processes[i].handle = 0;
        pid_generator = 0;
    }

    for (unsigned i = 0; i < NUM_CORES; i++)
    {
        core_state_t * s = &core_state[i];
        core_text_reloc[i] = 0;

        s->is_waiting_for_process = false;
        s->attached_pid = 0;
    }
    update_library_xml();
}

void gdb_on_core_setup(uint32_t core_id, uint32_t text_segment)
{
    gdb_lock_mutex(&mutex);
    //printf("GDB: Core %d is spawning at offset %#x\n", core_id, text_segment);
    core_text_reloc[core_id] = text_segment;
    update_library_xml();
    gdb_unlock_mutex(&mutex);
}

void gdb_on_process_created(dispatcher_handle_t handle, uint32_t segment_address)
{
    gdb_lock_mutex(&mutex);
    unsigned core_id = get_cpu_id();
    int empty_slot = -1;

    for (unsigned i = 0; i < MAX_PROCESS_ENTRIES; ++i)
    {
        if (processes[i].handle == 0)
        {
            empty_slot = i;
            break;
        }
    }

    if (empty_slot == -1)
    {
        // Process array is full
        gdb_unlock_mutex(&mutex);
        return;
    }
    process_t * entry = &processes[empty_slot];

    entry->handle = handle;
    entry->pid = ++pid_generator;
    entry->segment_address = segment_address;
    entry->core_id = core_id;

    struct dispatcher_shared_generic *disp = get_dispatcher_shared_generic(handle);
    string_copy((char *)entry->name, disp->name);
    update_library_xml();

    // printf("Process created name = %s, pid = %u, data = %u, text = %u\n", entry->name, entry->pid, data, text);
    for (unsigned i = 0; i < NUM_CORES; i++)
    {
        core_state_t * s = &core_state[i];
        if (s->is_waiting_for_process && s->attached_pid == entry->pid)
        {
            trap_dispatcher(handle);
            s->is_waiting_for_process = false;
        }
    }
    gdb_unlock_mutex(&mutex);
}

void gdb_on_process_destroyed(dispatcher_handle_t handle)
{
    gdb_lock_mutex(&mutex);
    // printf("Process destroyed %u\n", handle);
    for (unsigned i = 0; i < MAX_PROCESS_ENTRIES; ++i)
    {
        process_t * entry = &processes[i];
        if (entry->handle == handle)
        {
            entry->handle = 0;
        }
    }
    update_library_xml();
    gdb_unlock_mutex(&mutex);
}

void attach_process(uint32_t pid)
{
    gdb_lock_mutex(&mutex);
    unsigned core_id = get_cpu_id();
    core_state_t * s = &core_state[core_id];
    s->attached_pid = pid;
    dispatcher_handle_t handle = find_dispatcher(pid);
    if (handle != 0)
    {
        trap_dispatcher(handle);
        s->is_waiting_for_process = false;
    } else
    {
        s->is_waiting_for_process = true;
    }
    gdb_unlock_mutex(&mutex);
}

bool is_library_xml_updated(void)
{
    gdb_lock_mutex(&mutex);
    unsigned core_id = get_cpu_id();
    core_state_t * s = &core_state[core_id];
    bool result = s->library_xml_updated;
    gdb_unlock_mutex(&mutex);
    return result;
}

static dispatcher_handle_t find_dispatcher(uint32_t pid)
{
    for (unsigned i = 0; i < MAX_PROCESS_ENTRIES; ++i)
    {
        process_t * entry = &processes[i];
        if (entry->handle != 0 && entry->pid == pid)
        {
            return entry->handle;
        }
    }
    return 0;
}

static void trap_dispatcher(dispatcher_handle_t handle)
{
    struct dispatcher_shared_generic *disp = get_dispatcher_shared_generic(handle);
    if (disp->disabled)
    {
        arch_registers_state_t *disabled_area = dispatcher_get_disabled_save_area(handle);
        breakpoint_trap((uint32_t *) disabled_area->named.pc);
    } else
    {
        breakpoint_trap((uint32_t *) disp->dispatcher_run);
    }
}

uint32_t get_current_pid(void)
{
    gdb_lock_mutex(&mutex);
    unsigned cpu_id = get_cpu_id();
    uint32_t text_reloc = core_text_reloc[cpu_id];

    // Calculate current CPU's dcb_current by using the address of global variable in core 0 + kernel relocation offset.
    // This calculcation is needed as all cores uses the same GDB instance.
    // 1. Get the address of the global variable
    // 2. Add the text segment offset
    // 3. Cast it to a pointer to the dcb_current pointer.
    // 4. Dereference it to get the value of the global variable, which is a pointer to a dcb struct.
    struct dcb * current_disp = *((struct dcb **)(((uint32_t)&dcb_current) + text_reloc));
    
    if (current_disp == NULL)
    {
        gdb_unlock_mutex(&mutex);
        return 0;
    }
    dispatcher_handle_t handle = current_disp->disp;
    for (unsigned i = 0; i < MAX_PROCESS_ENTRIES; ++i)
    {
        process_t * entry = &processes[i];
        if (entry->handle == handle)
        {
            gdb_unlock_mutex(&mutex);
            return entry->pid;
        }
    }
    gdb_unlock_mutex(&mutex);
    return 42000;
}

void write_library_list(char * output, uint32_t offset, uint32_t length)
{
    gdb_lock_mutex(&mutex);

    *output++ = 'l';
    unsigned xml_length = string_length((char *)library_xml);
    if (offset < xml_length)
    {
        uint8_t * source = (uint8_t *) (library_xml + offset);
        uint8_t * dest = (uint8_t *) output;
        unsigned remaining = xml_length - offset;
        int min_length = remaining > length ? length : remaining;
        mem_copy(dest, source, min_length);
        output += min_length;
    }
    *output = 0;

    unsigned core_id = get_cpu_id();
    core_state_t * s = &core_state[core_id];
    s->library_xml_updated = false;

    gdb_unlock_mutex(&mutex);
}

void write_core_reloc(char * output)
{
    gdb_lock_mutex(&mutex);
    unsigned core_id = get_cpu_id();
    output = string_copy(output, "TextSeg=");
    output = int_to_hex(output, core_text_reloc[core_id]);
    gdb_unlock_mutex(&mutex);
}

void print_processes()
{
    gdb_lock_mutex(&mutex);
    char buffer[1024];
    char * pos = &buffer[0];
    pos = string_copy(pos, "== Processes ==\n");
    for (unsigned i = 0; i < MAX_PROCESS_ENTRIES; ++i)
    {
        process_t * entry = &processes[i];
        if (entry->handle != 0)
        {
            pos = string_copy(pos, "Name: ");
            pos = string_copy(pos, (char *)entry->name);
            pos = string_copy(pos, ", pid: ");
            pos = int_to_string(pos, entry->pid);        
            pos = string_copy(pos, ", segment start: 0x");
            pos = int_to_hex(pos, entry->segment_address);
            pos = string_copy(pos, ", core: ");
            pos = int_to_string(pos, entry->core_id);
            *pos++ = '\n';
        }
    }
    pos = string_copy(pos, "===============\n");
    gdb_print(buffer);
    gdb_unlock_mutex(&mutex);
}

static void update_library_xml(void)
{
    char * str = (char *) library_xml;
    str = string_copy(str, "<library-list>");
    for (unsigned i = 0; i < MAX_PROCESS_ENTRIES; ++i)
    {
        process_t * entry = &processes[i];
        if (entry->handle != 0)
        {
            str = string_copy(str, "<library name=\"");
            if (!matches((char *)entry->name, "shmac/sbin/"))
            {
                str = string_copy(str, "shmac/sbin/");
            }
            str = string_copy(str, (char *)entry->name);
            str = string_copy(str, "\"><segment address=\"0x");
            str = int_to_hex(str, entry->segment_address);
            str = string_copy(str, "\"/></library>");
        }
    }
    for (unsigned i = 0; i < NUM_CORES; i++)
    {
        core_state_t * s = &core_state[i];
        s->library_xml_updated = true;
    }
    str = string_copy(str, "</library-list>");
}

